#include <stdio.h>
#include <stdlib.h>
#include <msg.h>
#include <string.h>
#include <unistd.h>

#define MAXLEN 4096

struct cabuf {
	struct cabuf *next;
	size_t payloadlen;
	char sender_payload[0];
};

struct cabuf *cabufh;

static void cabuf_add(char *sender, void *buf, size_t len) {
	struct cabuf **scan;
	for(scan = &cabufh; (*scan) != NULL; scan = &((*scan)->next)) 
		;
	(*scan) = malloc(sizeof(struct cabuf) + len + strlen(sender) + 1);
	(*scan)->next = NULL;
	(*scan)->payloadlen = len;
	strcpy((*scan)->sender_payload, sender);
	memcpy((*scan)->sender_payload + strlen(sender) + 1, buf, len);
}

static ssize_t cabuf_get(char *sender, void *buf, size_t len) {
	struct cabuf **scan;
	for(scan = &cabufh; (*scan) != NULL; scan = &((*scan)->next)) {
		if (sender == NULL || *sender == 0 ||
				strcmp(sender, (*scan)->sender_payload) == 0) {
			struct cabuf *this = *scan;
			*scan = this->next;
			if (sender && *sender == 0) 
				strcpy(sender, this->sender_payload);
			if (this->payloadlen < len)
				len = this->payloadlen;
			memcpy(buf, this->sender_payload + strlen(this->sender_payload) + 1, len);
			free(this);
			return len; 
		}
	}
	return -1;
}

ssize_t a2ca_send(char *dest, const void *buf, size_t len) {
	return msg_send(dest, buf, len);
}

ssize_t a2ca_recv(char *sender, void *buf, size_t len) {
	ssize_t rval;
	msg_procname myname;
	msg_myname(myname);

	msg_send(myname, NULL, 0);
	while (1) {
		msg_procname msgsender;
		char msgbuf[MAXLEN];
		*msgsender = 0;
		size_t msglen = msg_recv(msgsender, msgbuf, MAXLEN);
		if (strcmp(msgsender,myname) == 0 && msglen == 0)
			break;
		else
			cabuf_add(msgsender, msgbuf, msglen);
	}

	rval = cabuf_get(sender, buf, len);
	return rval;
}

void a2ca_init(char *procname) {
	async_msg_init(procname);
}

void a2ca_fini(void) {
	msg_fini();
}
