#include <stdio.h>
#include <stdlib.h>
#include <string.h>
struct recq *head;

struct msgq {
	struct msgq *next;
	size_t sender_payloadlen;
	char sender_payload[0];
};

struct recq {
	struct recq *next;
	struct msgq *msgh;
	char *w4sender;
	char name[0];
};

#if 0
/* debug: print the data structure */
void printmsgq(void) {
	struct recq *this;
	for (this = head; this != NULL;this = this->next) {
		struct msgq *scan;
		printf("%s: waiting %s\n", this->name, this->w4sender);
		for(scan = this->msgh; scan != NULL; scan = scan->next) 
			printf("   msg from %s len %d\n", scan->sender_payload, scan->sender_payloadlen);
	}
}
#endif

struct recq **recp(char *receiver) {
	struct recq **scan;
	for (scan = &head; *scan != NULL; scan = &((*scan)->next))
		if (strcmp((*scan)->name,receiver) == 0)
			break;

	if ((*scan) == NULL &&
			((*scan) = malloc(sizeof(struct recq) + strlen(receiver) + 1)) != NULL) {
		struct recq *this = *scan;
		this->next = NULL;
		this->msgh = NULL;
		this->w4sender = NULL;
		strcpy(this->name, receiver);
	} 
	return scan;
}

void recp_clean(struct recq **scan) {
	struct recq *this = *scan;
	if (this->msgh == NULL && this->w4sender == NULL) {
		(*scan) = this->next;
		free(this);
	}
}

/* add a message in the data structure */
void msgq_add(char *sender, char *receiver, char *payload, size_t payloadlen,
	void (*send1msg)(char *receiver, char* sender_payload, size_t sender_payloadlen)) {

	struct recq **scan = recp(receiver);
	struct recq *this = *scan;
	size_t senderlen = strlen(sender) + 1;
	size_t sender_payloadlen = senderlen + payloadlen;
	//printf("adding %s %s %d\n",sender,receiver,payloadlen);
	if (this) {
		if (this->w4sender != NULL && 
				(this->w4sender[0] == 0 || strcmp(this->w4sender, sender) == 0)) {
			/* the receiver is waiting for this message so deliver it directly */
			char sender_payload[sender_payloadlen];
			strcpy(sender_payload, sender);
			memcpy(sender_payload + senderlen, payload, payloadlen);
			//printf("%s was waiting: direct delivery from %s len %d\n",receiver, sender_payload, payloadlen);
			send1msg(receiver, sender_payload, sender_payloadlen);
			free(this->w4sender);
			this->w4sender = NULL;
		} else {
			/* create a msg element and add it at the end of the queue */
			struct msgq *newmsg = malloc(sizeof(struct msgq) + sender_payloadlen);
			if (newmsg) {
				struct msgq **msgscan = &this->msgh;

				for (msgscan =  &this->msgh; *msgscan != NULL; msgscan = &((*msgscan)->next)) 
					;

				//printf("enqueueing\n");
				strcpy(newmsg->sender_payload, sender);
				memcpy(newmsg->sender_payload + senderlen, payload, sender_payloadlen);
				newmsg->sender_payloadlen = sender_payloadlen;
				newmsg->next = NULL;
				*msgscan = newmsg;
			}
		}
		recp_clean(scan);
	}
}

/* try to get a msg from the data structure */
void msgq_get(char *receiver, char *sender,
	void (*send1msg)(char *receiver, char* sender_payload, size_t sender_payloadlen)) {

	struct recq **scan = recp(receiver);
	struct recq *this = *scan;

	printf("Get %s from %s\n", receiver, sender);
	if (this) {
		struct msgq **msgscan = &this->msgh;

		for (msgscan =  &this->msgh; *msgscan != NULL; msgscan = &((*msgscan)->next)) 
			if (sender[0] == 0 || 
					strcmp(sender, (*msgscan)->sender_payload) == 0)
				break;

		if (*msgscan != NULL) {
			/* there is a msg matching the desired sender */
			struct msgq *thismsg = *msgscan;
			//printf("%s: delivery from %s len %d\n",receiver, thismsg->sender_payload, thismsg->sender_payloadlen);
			send1msg(receiver, thismsg->sender_payload, thismsg->sender_payloadlen);
			*msgscan = thismsg->next;
			free(thismsg);
		} else {
			/* no matching messages found: keep note of the sender required */
			//printf("register %s waiting for %s\n", receiver, sender);
			this->w4sender = strdup(sender);
		}
		recp_clean(scan);
	}
}


