#include<msg.h>
#include<string.h>

#define MAXLEN 4096
#define SERVER "S2AD"

/* async send: send 'S'+dest+payload to the server */
ssize_t s2a_send(char *dest, const void *buf, size_t len) {
	size_t preamblelen = strlen(dest) + 2;
	char outbuf[preamblelen + len];
	ssize_t rval;
	outbuf[0] = 'S';
	strcpy(outbuf+1, dest);
	memcpy(outbuf + preamblelen, buf, len);
	rval = msg_send(SERVER, outbuf, preamblelen + len);
	if (rval > preamblelen)
		rval -= preamblelen;
	return rval;
}

/* async receive: send 'R'+sender to the server,
	 then wait for a sender+payload */
ssize_t s2a_recv(char *sender, void *buf, size_t len) {
	char iobuf[MAXLEN];
	ssize_t rval;
	size_t rsenderlen;
	iobuf[0] = 'R';
	strcpy(iobuf+1, sender ? sender : "");
	msg_send(SERVER, iobuf, strlen(iobuf) + 1);
	rval = msg_recv(SERVER, iobuf, MAXLEN);
	if (rval > 0) {
		rsenderlen	= strlen(iobuf) + 1;
		if (sender && sender[0] == 0)
			strcpy(sender, iobuf);
		rval -= rsenderlen;
		if (len < rval) rval = len;
		if (rval > 0)
			memcpy(buf, iobuf + rsenderlen, len);
	}
	return rval;
}

void s2a_init(char *procname) {
	sync_msg_init(procname);
}

void s2a_fini(void) {
	msg_fini();
}

